import math
import xarray as xr
from DHDL import subprogs
from DHDL import subprogs_xarray

#########################################################
def change_input_ver_cross(slon,slat,dis,rot,lon_per_1km):
    elon = slon + dis * lon_per_1km * math.cos(math.radians(rot))  
    elat  = slat +  dis * 0.009 * math.sin(math.radians(rot))  

    return elon,elat
#######################################################

def check_vertical_coords(data):

    if not "ZPH" in data.variables:
        if "LEV" in data.variables:
            if data["LEV"].long_name == "Pressure level":
                ztype = 2
            else:
                ztype = 1
    else:        
        ztype = 0

    return ztype

#######################################################
def check_domain(vert_info,data):
    slon = vert_info[0]
    slat = vert_info[1]
    elon = vert_info[2]
    elat = vert_info[3]
    
    dimnameT,dimnameZ,dimnameY,dimnameX =subprogs.set_dim_name(data)
    LAT = data[dimnameY]
    LON = data[dimnameX]
    NX = len(LON)
    NY = len(LAT)
    
    if slon<LON[0] or elon>LON[NX-1] or slat<LAT[0] or elat>LAT[NY-1]:
        print (" vertical cross section is out of domain")
        exit(1)

    
#######################################################
def get3D(args,dimnameT,data,varname):
    
    if args.timenum is None and args.TIME is None:
        print("chose your time from following list \n")
        for i, a in enumerate(data[dimnameT].values):
            print(i,a)
        print("enter TNUM:")
        t = int(input())
        TTT = data[dimnameT].values[t]
        timesel={dimnameT : t}
        V3D = data[varname].isel(timesel)

        if args.WIND is not None:
            windU3D = data["U"].isel(timesel)
            windV3D = data["V"].isel(timesel)
            try:
                windW3D = data["W"].isel(timesel)
            except KeyError:
                windW3D = data["OMEGA"].isel(timesel)
                windW3D = -windW3D
                
    elif args.timenum is None:
        print("YYYY-MM-DD_HH:MN:SC format \n")
        ptn = r"\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d"
        if not re.match(ptn,args.TIME):
            print("TIME INPUT FORMAT WRONG")
            print("Y4-M2-D2TH2:M2:S2")
            print("ex 2020-07-06T03:00:00")
            exit(1)
        TTT = args.TIME
        timesel={dimnameT : args.TIME}
        V3D = data[varname].sel(timesel)
        if args.WIND is not None:
            windU3D = data["U"].sel(timesel)
            windV3D = data["V"].sel(timesel)
            try:
                windW3D = data["W"].sel(timesel)
            except KeyError:
                windW3D = data["OMEGA"].sel(timesel)
                windW3D = -windW3D 
    else:
        
        t = int(args.timenum)
        timesel={dimnameT : t}
        TTT = data[dimnameT].values[t]
        V3D = data[varname].isel(timesel)
        if args.WIND is not None:
            windU3D = data["U"].isel(timesel)
            windV3D = data["V"].isel(timesel)
            try:
                windW3D = data["W"].isel(timesel)
            except KeyError:
                windW3D = data["OMEGA"].isel(timesel)
                windW3D = -windW3D
                
    if args.WIND is not None:
        return V3D,TTT,windU3D,windV3D,windW3D
    else:
        return V3D,TTT
######################################################

def cross_section_xy(LON,LAT,lon0,lat0,lon1,lat1,lon_per_1km):
    
    dx = LON[2]-LON[1]
    dy = LAT[2]-LAT[1]
    dn = min(dx/lon_per_1km,dy/0.009)
    NX = len(LON)
    NY = len(LAT)

    distance = math.sqrt((lon1-lon0)*(lon1-lon0)/lon_per_1km/lon_per_1km \
                         +(lat1-lat0)*(lat1-lat0)/0.009/0.009)

    n = int(distance /dn)
    dis = distance/n

    unit_dx = (lon1-lon0)/n
    unit_dy = (lat1-lat0)/n
    n = n+1

    xy =[[0]*2 for i in range(n)]  ## initilization

    for i in range(n):
        x = lon0 + unit_dx * i
        y = lat0 + unit_dy * i
        xy[i] = (x,y)
        
    return xy
###################################################
def find_location(LON,xx):    
    i0 = -1
    dx = 0.0
    for i,lon in enumerate(LON):
        if lon > xx:
            dx = (LON[i] - xx)/(LON[1]-LON[0])
            i0 = i-1
            break
    return (i0,dx)
###################################################
def get_vertical_profile(V3D,i0,j0,dx,dy,missing_value):

    NZ            = len(V3D)    
    profile =[0 for i in range(NZ)]  ## initilization

    for k in range(NZ):
        
        a = V3D[k][j0][i0].values
        b = V3D[k][j0][i0+1].values
        c = V3D[k][j0+1][i0].values
        d = V3D[k][j0+1][i0+1].values
    
        if a > missing_value+1 and  b > missing_value+1 and \
           c > missing_value+1 and  d > missing_value+1:
            var= (1-dy)*(a*(1-dx)+b*dx)+ dy* (c*(1-dx)+d*dx)
            profile[k] = var
        else:
            profile[k] = missing_value
        
    return profile
#####################################################

###################################################
def get_ground_height(TOPO,i0,j0,dx,dy):
        
    a = TOPO[j0][i0].values
    b = TOPO[j0][i0+1].values
    c = TOPO[j0+1][i0].values
    d = TOPO[j0+1][i0+1].values
    ground_height= (1-dy)*(a*(1-dx)+b*dx)+ dy* (c*(1-dx)+d*dx)
    return ground_height
#####################################################

def vertical_interp(data_profile,zph_profile,LEV2,missing_value):

    NZ   = len(data_profile)
    NZ2 = len(LEV2)
    profile_new =[missing_value for i in range(NZ2)]  ## initilization

    for k2, zz2  in enumerate(LEV2):

        for k,zz in enumerate(zph_profile):

            if k==NZ :
                break

            if  zz2 >= zph_profile[k] and zz2 < zph_profile[k+1]:
                dz = (zz2 - zz)/(zph_profile[k+1]-zph_profile[k])

                if  data_profile[k] > missing_value+1 and \
                    data_profile[k+1] > missing_value+1 :
                    profile_new[k2] = (1-dz)*data_profile[k] + \
                                      dz*data_profile[k+1]
                else :
                    profile_new[k2] = missing_value
                
                break

    return profile_new
##########################################################
def calc_wind_comp_along_line(upro,vpro,theta,missing_value):
    M = len(upro)
    NZ= len(upro[0])
    along =[[0]*NZ for i in range(M)]  ## initilization    

    for m in range(M):
        for k in range(NZ):
            if upro[m][k] > missing_value+1:
                along[m][k] = upro[m][k]*math.cos(math.radians(theta))+\
                              vpro[m][k]*math.sin(math.radians(theta))
            else:
                along[m][k] = missing_value

    return  along
###########################################################
def make_vcross_sec(vert_info,data,varname,args,need_vertical_interp,TOPO):
    slon_ver = vert_info[0]
    slat_ver = vert_info[1]
    elon_ver = vert_info[2]
    elat_ver = vert_info[3]
    dlon     = vert_info[4]
    theta    = vert_info[5]
    #############################################################
    # check the dimension name    
    dimnameT,dimnameZ,dimnameY,dimnameX =subprogs.set_dim_name(data)
    #############################################################

    if isinstance(TOPO,xr.DataArray):
        optTOPO = 1
    else:
        optTOPO = 0
    
    if args.WIND is not None:
        V3D,TTT,winU3D,winV3D,winW3D  = get3D(args,dimnameT,data,varname)
    else:
        V3D,TTT  = get3D(args,dimnameT,data,varname)

    ## GET missing value #################
    missing_value = V3D.encoding["missing_value"]
    #############################################################
    # to change unit (especially, for QR,QV, kg/kg -> g/kg)
    if varname == "QV" or varname == "QR": 
        V3D = V3D * 1000.0
        data[varname].attrs['units'] = "g/kg"
    #############################################################    
    
    ### check vertical coordinate (for CReSS, ZPH is used)
    ztype = check_vertical_coords(data)
    
    if ztype == 0 and need_vertical_interp == "YES": # cress
        print("cress case, height above sea level coordinate \n")
        Z3D = data["ZPH"]
        LEV = data[dimnameZ].values
        NZ = len(LEV)        
        dz_new = LEV[1] - LEV[0]
        bottom_new = 0.0
        top_new = int((LEV[NZ-1]-bottom_new)/dz_new)*dz_new +bottom_new
        NZ2  = int((top_new-bottom_new)/dz_new) + 1
        Z1D =[bottom_new+dz_new*i  for i in range(NZ2)]

    elif ztype == 0 and need_vertical_interp == "NO": # cress
        print("cress case, terrain-following coordinate \n")
        Z1D = data[dimnameZ].values
        NZ  = len(Z1D)

        if optTOPO == 1:
            print("In CReSS, originally terrain-following ");
            print("Topography mask can not be applied")
            print("when vertical interpolation is applied ");
            exit(1)
            
    elif ztype == 1 :
        print("simple vertical coordinate \n")
        Z1D = data[dimnameZ].values
        NZ  = len(Z1D)

        if need_vertical_interp == "YES":
            print("ztype=1 case, vertical interpolation is not necessary\n ");
            exit(1)

        if optTOPO == 1:
            print("ztype=1 case, topography mask is not available\n")
            exit(1)
            
    elif ztype == 2 :

        if need_vertical_interp == "YES":
            print("ztype=2 case, vertically interp will make z-coordinate")
            timesel = {dimnameT : TTT}
            Z3D = data["Z"].sel(timesel)
            LEV = data[dimnameZ].values
            NZ = len(LEV)        
            dz_new     = 100.0
            bottom_new = 0.0
            top_new    = 14000.0
            NZ2  = int((top_new-bottom_new)/dz_new) + 1
            Z1D =[bottom_new+dz_new*i  for i in range(NZ2)]
        else:
            print("ztype=2 case, originally p-coordinate")
            Z1D = data[dimnameZ].values
            NZ  = len(Z1D)
            
            if optTOPO == 1:
                timesel = {dimnameT : TTT}
                Z3D = data["Z"].sel(timesel)

    LAT = data[dimnameY].values
    LON = data[dimnameX].values
    xy = cross_section_xy(LON,LAT,slon_ver,slat_ver,elon_ver,elat_ver,dlon)
    M = len(xy)
    profile =[[0.e0]*NZ for i in range(M)]  ## initilization

    if args.WIND is not None:
        upro =[[0.e0]*NZ for i in range(M)]  ## initilization
        vpro =[[0.e0]*NZ for i in range(M)]  ## initilization
        wpro =[[0.e0]*NZ for i in range(M)]  ## initilization

    if need_vertical_interp == "YES" or TOPO != "NONE":
        zph_profile    =[[0.e0]*NZ for i in range(M)]   ## initilization

    ##############################################
    if need_vertical_interp == "YES":
        interp_profile =[[0.e0]*NZ2 for i in range(M)]  ## initilization

        if args.WIND is not None:
            uintp =[[0.e0]*NZ2 for i in range(M)]  ## initilization
            vintp =[[0.e0]*NZ2 for i in range(M)]  ## initilization
            wintp =[[0.e0]*NZ2 for i in range(M)]  ## initilization
    ##############################################
    
    ##############################################
    # get vertical profile at a grid point along the line
    for m, (xx, yy)  in enumerate(xy):
        (i0,dx)= find_location(LON,xx)
        (j0,dy)= find_location(LAT,yy)
        profile[m] = get_vertical_profile(V3D,i0,j0,dx,dy,missing_value)

        if args.WIND is not None:
            upro[m] = get_vertical_profile(winU3D,i0,j0,dx,dy,\
                                           missing_value)
            vpro[m] = get_vertical_profile(winV3D,i0,j0,dx,dy,\
                                           missing_value)
            wpro[m] = get_vertical_profile(winW3D,i0,j0,dx,dy,\
                                           missing_value)

        if need_vertical_interp == "YES" or TOPO != "NONE":
            zph_profile[m] = get_vertical_profile(Z3D,i0,j0,dx,dy,\
                                                  missing_value)

        if need_vertical_interp == "YES":
            # get vertically interpotation 
            interp_profile[m] = vertical_interp(profile[m],\
                                                zph_profile[m],\
                                                Z1D,missing_value)
            
            if args.WIND is not None:
                uintp[m] = vertical_interp(upro[m],zph_profile[m],\
                                           Z1D,missing_value)
                vintp[m] = vertical_interp(vpro[m],zph_profile[m],\
                                           Z1D,missing_value)
                wintp[m] = vertical_interp(wpro[m],zph_profile[m],\
                                           Z1D,missing_value)

        if optTOPO == 1:
            groundheight = get_ground_height(TOPO,i0,j0,dx,dy)
            if need_vertical_interp == "NO":
                for k, z  in enumerate(zph_profile[m]):
                    if z < groundheight:
                        profile[m][k] = missing_value

                if args.WIND is not None:
                    for k, z  in enumerate(zph_profile[m]):
                        if z < groundheight:
                            upro[m][k] = missing_value
                            vpro[m][k] = missing_value
                            wpro[m][k] = missing_value
            else:
                for k, z  in enumerate(Z1D):
                    if z < groundheight:
                        interp_profile[m][k] = missing_value
                if args.WIND is not None:
                    for k, z  in enumerate(Z1D):
                        if z < groundheight:
                            uintp[m][k] = missing_value
                            vintp[m][k] = missing_value
                            wintp[m][k] = missing_value
                    
    # end of loop for m ###########################

    
    # calcuate wind component parallel to the vertical cross section
    if args.WIND is not None:
        if need_vertical_interp == "YES":
            uv_along= calc_wind_comp_along_line(uintp,vintp,theta,\
                                                missing_value)
        else :
            uv_along= calc_wind_comp_along_line(upro,vpro,theta,\
                                                missing_value)

    ###### make xarray.Dataset  ################
    if need_vertical_interp == "YES":
        VER = subprogs_xarray.make_profile_xarray(interp_profile,Z1D,xy,dlon,\
                                                  missing_value)
        if args.WIND is not None:
            UVALONG= subprogs_xarray.make_profile_xarray(uv_along,Z1D,xy,dlon,\
                                                         missing_value)
            WVER =subprogs_xarray.make_profile_xarray(wintp,Z1D,xy,dlon,\
                                                      missing_value)

    else:
        VER = subprogs_xarray.make_profile_xarray(profile,Z1D,xy,dlon,\
                                                  missing_value)
        if args.WIND is not None:
            UVALONG= subprogs_xarray.make_profile_xarray(uv_along,Z1D,xy,dlon,\
                                                         missing_value)
            WVER =subprogs_xarray.make_profile_xarray(wpro,Z1D,xy,dlon,\
                                                      missing_value)

    # data thin out for vector drawing
    if args.WINDTHIN is not None:
        if(int(args.WINDTHIN)==0):
            print("use must set WINDTHIN>1")
            exit(1)
            
        UVALONG = subprogs_xarray.mabiki2D_xarray(UVALONG,\
                                                int(args.WINDTHIN))
        WVER    = subprogs_xarray.mabiki2D_xarray(WVER,\
                                                int(args.WINDTHIN))

            
    colormap,clevels =subprogs.set_color_range(args,VER)

    if args.WIND is not None:
        scaleUV,maxUV = subprogs.optimize_wind_vector_length(UVALONG,WVER)

    if args.WIND is not None:
        return Z1D,VER,TTT,colormap,clevels,\
            UVALONG,WVER,scaleUV,maxUV
    else:
        return Z1D,VER,TTT,colormap,clevels

    ###############################################################

    
